#include <stdlib.h>
#include "gsl_interp.h"

gsl_interp_accel *
gsl_interp_accel_alloc (void)
{
  gsl_interp_accel *a = (gsl_interp_accel *) malloc (sizeof (gsl_interp_accel));
  if (a == 0)
    {
      return NULL;
    }

  a->cache = 0;
  a->hit_count = 0;
  a->miss_count = 0;

  return a;
}

int
gsl_interp_accel_reset (gsl_interp_accel * a)
{
  a->cache = 0;
  a->hit_count = 0;
  a->miss_count = 0;

  return GSL_SUCCESS;
}

void
gsl_interp_accel_free (gsl_interp_accel * a)
{
  if (a==NULL) return;
  free (a);
}

/* Perform a binary search of an array of values.
 *
 * The parameters index_lo and index_hi provide an initial bracket,
 * and it is assumed that index_lo < index_hi. The resulting index
 * is guaranteed to be strictly less than index_hi and greater than
 * or equal to index_lo, so that the implicit bracket [index, index+1]
 * always corresponds to a region within the implicit value range of
 * the value array.
 *
 * Note that this means the relationship of 'x' to x_array[index]
 * and x_array[index+1] depends on the result region, i.e. the
 * behaviour at the boundaries may not correspond to what you
 * expect. We have the following complete specification of the
 * behaviour.
 * Suppose the input is x_array[] = { x0, x1, ..., xN }
 *    if ( x == x0 )           then  index == 0
 *    if ( x > x0 && x <= x1 ) then  index == 0, and sim. for other interior pts
 *    if ( x == xN )           then  index == N-1
 *    if ( x > xN )            then  index == N-1
 *    if ( x < x0 )            then  index == 0
 */

size_t gsl_interp_bsearch(const double x_array[], double x,
                   size_t index_lo, size_t index_hi)
{
    size_t ilo = index_lo;
    size_t ihi = index_hi;
    while(ihi > ilo + 1) {
        size_t i = (ihi + ilo)/2;
        if(x_array[i] > x)
            ihi = i;
        else
            ilo = i;
    }
    
    return ilo;
}

size_t gsl_interp_accel_find(gsl_interp_accel * a, const double xa[], size_t len, double x)
{
    size_t x_index = a->cache;

    if(x < xa[x_index]) {
        a->miss_count++;
        a->cache = gsl_interp_bsearch(xa, x, 0, x_index);
    }
    else if(x >= xa[x_index + 1]) {
        a->miss_count++;
        a->cache = gsl_interp_bsearch(xa, x, x_index, len-1);
    }
    else {
        a->hit_count++;
    }
    
    return a->cache;
}

